-- Diamond Lite 1.1 by Mike Nelson

-- A simplified version of Diamond 1.0
-- A Java-like class system for Euphoria with single inheritance. 
-- Exception handling for errors.  Polymorphic methods.
-- v1.1                               
-- Added null class to conform to Diamond 1.1, added this_class()
-- v1.1.1
-- Corrections to conform to Diamond.

without trace
without warning
-- CONSTANTS:
-- Entity marker:
constant MARKER=-power(2,30)
-- Methods:
global constant NULL_METHOD=-2
constant BAD_METHOD=-1
-- Meaningless return values:
global constant NIL=0
-- Booleans:
global constant FALSE=0
global constant TRUE=1
-- Class types:
-- Property and method varieties:
global constant INSTANCE=1
global constant CLASS=2
-- Exception:
constant EXCEPTION=3 
-- Null class:
constant NULL_CLASS=4
-- Program contexts:
constant CLASS_DEFINITION=5
constant MAIN_PROGRAM=6
constant INSTANCE_METHOD=7
constant CLASS_METHOD=8                          
-- Empty lists:
global constant NONE={}
-- Program stack:
constant STACK_EXPAND=repeat(NIL,16)
-- Identifiers:
constant ID_INIT="ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz"
constant ID_DIGIT="0123456789"
constant ID_CHAR=ID_INIT&ID_DIGIT&'_'
-- Predefined enitities:
global constant Entity={1,0,MARKER}
global constant Exception={2,0,MARKER}
global constant Null_Class={3,0,MARKER}
global constant Null_Instance={3,1,MARKER}
-- Validation:
constant VALID_ENTITY_TYPE={INSTANCE,CLASS}
-- Error messages:
constant CONSTANT_NAME={"instance","class","exception","null class",
"class definition","main program","instance method","class method"}
constant CONSTANT_CAP_NAME={"Instance","Class","Exception","Null class",
"Class definition","Main program","Instance method","Class method"}
-- Misc.:
constant ADD_NONE={NONE}
constant TRIPLE_NONE=repeat(NONE,3)

-- VARIABLES:
-- Trash can for unwanted return values:
integer SETVOIDID

type void_object(object x)
   if not equal(x,NIL) then call_proc(SETVOIDID,{}) end if
   return TRUE
end type    

global void_object VOID

procedure setvoid()
   VOID=NIL
end procedure

SETVOIDID=routine_id("setvoid")

-- Class database:
sequence ClassName ClassName={"Entity","Exception","Null_Class"} 
sequence ClassType ClassType={CLASS,EXCEPTION,NULL_CLASS}
sequence SuperclassList SuperclassList={{1},{2},NONE}
sequence InstancePropertyName InstancePropertyName=TRIPLE_NONE
sequence InstancePropertySource InstancePropertySource=TRIPLE_NONE
sequence InstancePropertyDefault InstancePropertyDefault=TRIPLE_NONE
sequence ClassPropertyName ClassPropertyName=TRIPLE_NONE
sequence ClassPropertySource ClassPropertySource=TRIPLE_NONE
sequence ClassPropertyDefault ClassPropertyDefault=TRIPLE_NONE
sequence InstanceMethodName InstanceMethodName=TRIPLE_NONE
sequence InstanceMethodSource InstanceMethodSource=TRIPLE_NONE
sequence InstanceMethodID InstanceMethodID=TRIPLE_NONE
sequence ClassMethodName ClassMethodName=TRIPLE_NONE
sequence ClassMethodSource ClassMethodSource=TRIPLE_NONE
sequence ClassMethodID ClassMethodID=TRIPLE_NONE

-- Entity database:
integer InstanceCounter InstanceCounter=1
sequence InstanceEntity InstanceEntity={NONE}&repeat(NIL,127) 
integer ClassCounter ClassCounter=0
sequence ClassEntity ClassEntity=TRIPLE_NONE
-- Program stack:
integer ProgramCounter ProgramCounter=1
sequence Context Context=STACK_EXPAND Context[1]=MAIN_PROGRAM
sequence Caller Caller=STACK_EXPAND
sequence This This=STACK_EXPAND
sequence ThisMethod ThisMethod=STACK_EXPAND
-- Exceptions:
integer PendingException PendingException=NIL
integer CaughtException CaughtException=NIL 
-- Error handling:
integer ErrorScreenWidth ErrorScreenWidth=80

-- ROUTINES:
-- Error handling:
function get_class_name(integer index)
   return ClassName[index]  
end function

function context_string()
   integer context,thisClass
   sequence cs
   if PendingException!=NIL then
      cs="With "&get_class_name(PendingException)&" exception pending.  "
   elsif CaughtException!=NIL then
      cs="With "&get_class_name(CaughtException)&" exception caught.  "
   else
      cs=""
   end if
   for i=ProgramCounter to 1 by -1 do
      if i=ProgramCounter then cs&="In " else cs&=", called from " end if
      context=Context[i]
      if context!=MAIN_PROGRAM then
         thisClass=Caller[i]
         cs&=get_class_name(thisClass)&" "
      end if
      cs&=CONSTANT_NAME[context]
      if context=INSTANCE_METHOD then
         cs&=" "&InstanceMethodName[thisClass][ThisMethod[i]]
      end if
      if context=CLASS_METHOD then
         cs&=" "&ClassMethodName[thisClass][ThisMethod[i]]
      end if
   end for
   return cs&"."
end function

type message_string(object x)
    object char
    if atom(x) then return FALSE end if
    for i=1 to length(x) do
        char=x[i]
        if not integer(char) then return FALSE end if
        if char<32 then return FALSE end if
        if char>255 then return FALSE end if
    end for
    return TRUE
end type  

global procedure error_screen_width(object width)
   if not integer(width) then return end if
   if width<20 then return end if
   ErrorScreenWidth=width
end procedure

global procedure fatal_error(object ErrorMessage)
   integer fn,len,offset
   if not message_string(ErrorMessage) then
      ErrorMessage="*** Invalid error message. ***"
   end if
   ErrorMessage="FATAL ERROR:  "&ErrorMessage&"  "&context_string()
   len=length(ErrorMessage)
   offset=0
   while len>ErrorScreenWidth-1 do
      for i=ErrorScreenWidth-1 to 1 by -1 do
         if ErrorMessage[i+offset]=' ' then
            ErrorMessage[i+offset]='\n'
            offset+=i
            len-=i
            exit
         end if
      end for
   end while
   fn=open("dlite.err","w")
   if fn!=-1 then puts(fn,ErrorMessage) end if
   machine_proc(37,ErrorMessage) -- crash_message
   ? 1/0 -- to generate ex.err
end procedure                        

-- Program stack:
procedure expand_stack()
   Context&=STACK_EXPAND
   Caller&=STACK_EXPAND
   This&=STACK_EXPAND
   ThisMethod&=STACK_EXPAND
end procedure

-- The identifier type:
global type identifier(object x)
   if atom(x) then return FALSE end if
   if length(x)=0 then return FALSE end if
   if not find(x[1],ID_INIT) then return FALSE end if
   for i=2 to length(x) do
      if not find(x[i],ID_CHAR) then return FALSE end if
   end for
   return TRUE
end type

type method_specifier(object x)
   object x2
   if atom(x) then return FALSE end if
   if length(x)!=2 then return FALSE end if
   if not identifier(x[1]) then return FALSE end if
   x2=x[2]
   if not integer(x2) then return FALSE end if
   return x2>=0
end type 

type method_name(object x)
   integer char,index,len,zflag
   if atom(x) then return FALSE end if
   index=find('#',x)
   if index=NIL then return FALSE end if
   len=length(x)
   if index=len then return FALSE end if
   if not identifier(x[1..index-1]) then return FALSE end if
   zflag=FALSE
   for i=index+1 to len do
      char=x[1]
      if not find(x[i],ID_DIGIT) then return FALSE end if   
      if x[i]='0' then
         if zflag=FALSE then 
            if i!=len then return FALSE end if
         end if
      else
         zflag=TRUE
      end if
   end for
   return TRUE
end type

-- The entity type:
global type entity(object x)
   object xi
   if atom(x) then return FALSE end if
   if length(x)!=3 then return FALSE end if
   if not equal(x[3],MARKER) then return FALSE end if
   xi=x[1]
   if not integer(xi) then return FALSE end if
   if xi<1 then return FALSE end if
   if xi>ClassCounter then return FALSE end if
   xi=x[2]
   if not integer(xi) then return FALSE end if
   if xi<0 then return FALSE end if
   return xi<=InstanceCounter
end type

global type instance_entity(object x)
   if not entity(x) then return FALSE end if
   return x[2]!=0
end type

global type deleted_instance(object x)
   if not instance_entity(x) then return FALSE end if
   return atom(InstanceEntity[x[2]])
end type

global type class_entity(object x)
   if not entity(x) then return FALSE end if
   return x[2]=0
end type  


-- Class information:
global function get_class(object theEntity)
   if not entity(theEntity) then return NONE end if
   theEntity[2]=0
   return theEntity
end function  

global function class_name(object theEntity)       
   if not entity(theEntity) then return NONE end if
   return get_class_name(theEntity[1])
end function  

global function same_class(object entity1,object entity2)
   if not entity(entity1) then return FALSE end if
   if not entity(entity2) then return FALSE end if
   return entity1[1]=entity2[1]
end function

global function extends(object entity1,object entity2)
   integer class1
   if not entity(entity1) then return FALSE end if
   if not entity(entity2) then return FALSE end if
   class1=entity1[1]
   if class1=3 then return TRUE end if
   return find(entity2[1],SuperclassList[class1])!=0
end function

-- Class operations:
global function this()
   if Context[ProgramCounter]<INSTANCE_METHOD then
      fatal_error("In this().  Not allowed.")
   end if
   return This[ProgramCounter]
end function  

global function this_class()  
   sequence thisEntity 
   if Context[ProgramCounter]<INSTANCE_METHOD then
      fatal_error("In this_class().  Not allowed.")
   end if
   thisEntity=This[ProgramCounter]
   thisEntity[2]=0
   return thisEntity
end function

global function super()   
   if Context[ProgramCounter]<INSTANCE_METHOD then
      fatal_error("In super().  Not allowed.")
   end if
   return MARKER
end function

global function get_property(object target,object prop)
   integer caller,context,flag,index,source,targetType
   integer thisClass,thisInstance
   sequence propName,temp
   object instanceValue 
   context=Context[ProgramCounter]
   if context<MAIN_PROGRAM then
      fatal_error("In get_property().  Not allowed.")
   end if
   if not entity(target) then
      fatal_error("In get_property().  Target is not an entity.")
   end if
   thisClass=target[1]
   thisInstance=target[2]
   if thisInstance!=0 then
      instanceValue=InstanceEntity[thisInstance]
      if atom(instanceValue) then
         fatal_error("In get_property().  Target is a deleted instance.")
      end if
      propName=InstancePropertyName[thisClass]
      targetType=INSTANCE
   else 
      if ClassType[thisClass]>CLASS then
         fatal_error("In get_property().  "
         &CONSTANT_CAP_NAME[ClassType[thisClass]]&" "
         &get_class_name(thisClass)&" not allowed as target.")    
      end if
      propName=ClassPropertyName[thisClass]
      targetType=CLASS
   end if
   index=find(prop,propName)
   if index=NIL then
      if identifier(prop) then
         fatal_error("In get_property().  Class "&get_class_name(thisClass)
         &" does not define "&CONSTANT_NAME[targetType]&" property "
         &prop&".")
      end if
      fatal_error("In get_property().  Invalid property name.")
   end if
   if targetType=INSTANCE then
      temp=InstancePropertySource[thisClass]
      source=temp[index]
   else
      temp=ClassPropertySource[thisClass]
      source=temp[index]
   end if
   flag=FALSE
   if context!=MAIN_PROGRAM then
      caller=Caller[ProgramCounter]
      if find(caller,SuperclassList[source]) then flag=TRUE end if
   end if
   if flag=FALSE then
      fatal_error("In get_property().  Access to "&get_class_name(thisClass)
      &" "&CONSTANT_NAME[targetType]&" property "&prop&" denied.")
   end if
   if thisInstance!=0 then
      return instanceValue[index]
   else
      return ClassEntity[thisClass][index]
   end if
end function

global procedure set_property(object target,object prop,object newValue)
   integer caller,context,flag,index,source
   integer targetType,thisClass,thisInstance
   sequence propName,temp
   context=Context[ProgramCounter]
   if context<MAIN_PROGRAM then
      fatal_error("In set_property().  Not allowed.")
   end if
   if not entity(target) then
      fatal_error("In set_property().  Target is not an entity.")
   end if
   thisClass=target[1]
   thisInstance=target[2]
   if thisInstance!=0 then
      if atom(InstanceEntity[thisInstance]) then
         fatal_error("In set_property().  Target is a deleted instance.")
      end if
      propName=InstancePropertyName[thisClass]
      targetType=INSTANCE
   else
      if ClassType[thisClass]>CLASS then
         fatal_error("In get_property().  "
         &CONSTANT_CAP_NAME[ClassType[thisClass]]&" "
         &get_class_name(thisClass)&" not allowed as target.")    
      end if
      propName=ClassPropertyName[thisClass]
      targetType=CLASS
   end if
   index=find(prop,propName)
   if index=NIL then
      if identifier(prop) then
         fatal_error("In set_property().  Class "&get_class_name(thisClass)
         &" does not define "&CONSTANT_NAME[targetType]&" property "
         &prop&".")
      end if
      fatal_error("In set_property().  Invalid property name.")
   end if
   if targetType=INSTANCE then 
      temp=InstancePropertySource[thisClass]
      source=temp[index]
   else 
      temp=ClassPropertySource[thisClass]
      source=temp[index]
   end if
   flag=FALSE
   if context!=MAIN_PROGRAM then
      caller=Caller[ProgramCounter]
      if find(caller,SuperclassList[source]) then flag=TRUE end if
   end if
   if flag=FALSE then
      fatal_error("In set_property().  Access to "&get_class_name(thisClass)
      &" "&CONSTANT_NAME[targetType]&" property "&prop&" denied.")
   end if
   if thisInstance!=0 then
      InstanceEntity[thisInstance][index]=newValue
   else
      ClassEntity[thisClass][index]=newValue
   end if
end procedure

global function call_method(object target,object meth,object Param)
   integer  context,ID,index,source,searchClass
   integer targetType,thisClass,thisInstance
   sequence methName,temp
   object methReturn
   context=Context[ProgramCounter]
   if context<MAIN_PROGRAM or PendingException!=NIL then
      fatal_error("In call_method().  Not allowed.")
   end if
   if entity(target) then                                   
      searchClass=target[1]
   elsif context!=MAIN_PROGRAM and equal(target,MARKER) then
      searchClass=SuperclassList[Caller[ProgramCounter]][2]
      target=This[ProgramCounter]
   else
      fatal_error("In call_method().  Target is not an entity.")
   end if                       
   if atom(Param) then
      fatal_error("In call_method().  Parameter list must be a sequence.")
   end if
   thisClass=target[1]
   thisInstance=target[2]
   if thisInstance!=0 then
      if atom(InstanceEntity[thisInstance]) then
         fatal_error("In call_method().  Target is a deleted instance.")
      end if
      methName=InstanceMethodName[searchClass]
      targetType=INSTANCE
   else
      if ClassType[searchClass]>CLASS then
         fatal_error("In call_method().  "
         &CONSTANT_CAP_NAME[ClassType[searchClass]]&" "
         &get_class_name(thisClass)&" not allowed as target.")    
      end if
      methName=ClassMethodName[searchClass]  
      targetType=CLASS
   end if
   meth&=sprintf("#%d",length(Param))
   index=find(meth,methName)
   if index=NIL then
      if method_name(meth) then
         fatal_error("In call_method().  Class "&get_class_name(searchClass)
         &" does not define "&CONSTANT_NAME[targetType]&" method "
         &meth&".")
      end if
      fatal_error("In call_method().  Invalid method name.")
   end if 
   if targetType=INSTANCE then 
      temp=InstanceMethodSource[searchClass]
      source=temp[index]
      temp=InstanceMethodID[searchClass]
      ID=temp[index]
   else           
      temp=ClassMethodSource[searchClass]
      source=temp[index]
      temp=ClassMethodID[searchClass]
      ID=temp[index]
   end if
   if ID=NULL_METHOD then return NIL end if
   ProgramCounter+=1
   if ProgramCounter>length(Context) then expand_stack() end if
   Context[ProgramCounter]=targetType+INSTANCE_METHOD-INSTANCE
   Caller[ProgramCounter]=source
   This[ProgramCounter]=target
   ThisMethod[ProgramCounter]=index
   methReturn=call_func(ID,Param)
   ProgramCounter-=1
   return methReturn
end function  


-- Class definition:  

procedure new_class(sequence Name,sequence from)
   integer index
   index=find(Name,ClassName)
   if index!=NIL then
      fatal_error("In "&from&"().  "&get_class_name(index)
      &" already defined.")   
   end if 
   ClassCounter+=1 
   ClassName&={Name}
   ClassType&=NIL
   SuperclassList&=ADD_NONE
   InstancePropertyName&=ADD_NONE
   InstancePropertySource&=ADD_NONE
   InstancePropertyDefault&=ADD_NONE
   ClassPropertyName&=ADD_NONE
   ClassPropertySource&=ADD_NONE
   ClassPropertyDefault&=ADD_NONE
   InstanceMethodName&=ADD_NONE
   InstanceMethodSource&=ADD_NONE
   InstanceMethodID&=ADD_NONE
   ClassMethodName&=ADD_NONE
   ClassMethodSource&=ADD_NONE
   ClassMethodID&=ADD_NONE
   ClassEntity&=ADD_NONE
   ProgramCounter+=1
   if ProgramCounter>length(Context) then expand_stack() end if
   Context[ProgramCounter]=CLASS_DEFINITION
   Caller[ProgramCounter]=ClassCounter
   This[ProgramCounter]={ClassCounter,0,MARKER}
   ThisMethod[ProgramCounter]=NIL
end procedure

global function class(object Name,object Extends)
   integer superClass
   if ProgramCounter!=1 or PendingException!=NIL then
      fatal_error("In class().  Not allowed.")
   end if
   if not identifier(Name) then
      fatal_error("In class().  Invalid class name.")
   end if
   new_class(Name,"class")
   ClassType[ClassCounter]=CLASS
   if not class_entity(Extends) then
      fatal_error("In class().  Invalid superclass.")
   end if
   superClass=Extends[1]
   if ClassType[superClass]!=CLASS then 
      fatal_error("In class().  Superclass "&get_class_name(superClass)
      &" is not a normal class.")
   end if
   SuperclassList[ClassCounter]=ClassCounter&SuperclassList[superClass]
   InstancePropertyName[ClassCounter]=InstancePropertyName[superClass]
   InstancePropertySource[ClassCounter]=InstancePropertySource[superClass]
   InstancePropertyDefault[ClassCounter]=InstancePropertyDefault[superClass]
   ClassPropertyName[ClassCounter]=ClassPropertyName[superClass]
   ClassPropertySource[ClassCounter]=ClassPropertySource[superClass]
   ClassPropertyDefault[ClassCounter]=ClassPropertyDefault[superClass]
   InstanceMethodName[ClassCounter]=InstanceMethodName[superClass]
   InstanceMethodSource[ClassCounter]=InstanceMethodSource[superClass]
   InstanceMethodID[ClassCounter]=InstanceMethodID[superClass]
   ClassMethodName[ClassCounter]=ClassMethodName[superClass]
   ClassMethodSource[ClassCounter]=ClassMethodSource[superClass]
   ClassMethodID[ClassCounter]=ClassMethodID[superClass]
   return This[2]
end function

global procedure end_class()
   if Context[ProgramCounter]!=CLASS_DEFINITION then
      fatal_error("In end_class().  Not allowed.")
   end if
   ClassEntity[ClassCounter]=ClassPropertyDefault[ClassCounter]
   ProgramCounter=1
end procedure

global function exception(object Name,object Extends)
   integer superClass
   if ProgramCounter!=1 or PendingException!=NIL then
      fatal_error("In exception().  Not allowed.")
   end if
   if not identifier(Name) then
      fatal_error("In exception().  Invalid exception name.")
   end if
   new_class(Name,"exception")
   if not class_entity(Extends) then
      fatal_error("In exception().  Invalid superclass.")
   end if
   superClass=Extends[1]
   if ClassType[superClass]!=EXCEPTION then
      fatal_error("In exception().  Superclass "&get_class_name(superClass)
      &" is not an exception.")
   end if
   ClassType[ClassCounter]=EXCEPTION
   SuperclassList[ClassCounter]=ClassCounter&SuperclassList[superClass]
   ProgramCounter=1
   return This[2]
end function

global procedure property(object Name,object entityType,object default)
   integer index
   if Context[ProgramCounter]!=CLASS_DEFINITION then
      fatal_error("In property().  Not allowed.")
   end if
   if not identifier(Name) then
      fatal_error("In property().  Invalid property name.")
   end if
   if not find(entityType,VALID_ENTITY_TYPE) then
      fatal_error("In property().  Property "&Name
      &" has invalid entity type.")
   end if
   if entityType=INSTANCE then
      index=find(Name,InstancePropertyName[ClassCounter])
      if index!=NIL then
         if InstancePropertySource[ClassCounter][index]=ClassCounter then
            fatal_error("In property().  Instance property "&Name
            &" has already been defined.")
         end if
         fatal_error("In property().  Attempt to override "
         &get_class_name(InstancePropertySource[ClassCounter][index])
         &" instance property "&Name&".")
      end if
   else
      index=find(Name,ClassPropertyName[ClassCounter])
      if index!=NIL then
         if ClassPropertySource[ClassCounter][index]=ClassCounter then
            fatal_error("In property().  Class property "&Name
            &" has already been defined.")
         end if
         fatal_error("In property().  Attempt to override "
         &get_class_name(ClassPropertySource[ClassCounter][index])
         &" class property "&Name&".")
      end if
   end if                            
   if entityType=INSTANCE then
      InstancePropertyName[ClassCounter]&={Name}
      InstancePropertySource[ClassCounter]&=ClassCounter
      InstancePropertyDefault[ClassCounter]&={default}
   else
      ClassPropertyName[ClassCounter]&={Name}
      ClassPropertySource[ClassCounter]&=ClassCounter
      ClassPropertyDefault[ClassCounter]&={default}
   end if
end procedure

global procedure method(object Name,object params,object entityType,
object ID)
   integer index
   sequence methSource
   if Context[ProgramCounter]!=CLASS_DEFINITION then
      fatal_error("In method().  Not allowed.")
   end if
   if not identifier(Name) then
      fatal_error("In method().  Invalid method name.")
   end if
   if not find(entityType,VALID_ENTITY_TYPE) then
      fatal_error("In method().  Method "&Name&" has invalid entity type.")
   end if
   if not integer(params) or params<0 then
      fatal_error("In method().  Method "&Name
      &" has invalid parameter count.")
   end if
   Name&=sprintf("#%d",params)
   if entityType=INSTANCE then
      index=find(Name,InstanceMethodName[ClassCounter])
   else
      index=find(Name,ClassMethodName[ClassCounter])
   end if
   if not integer(ID) or ID=BAD_METHOD or ID<NULL_METHOD then
      fatal_error("In method().  "&CONSTANT_CAP_NAME[entityType]
      &" method "&Name&" has invalid routine id.")
   end if
   if index=NIL then
      if entityType=INSTANCE then
         InstanceMethodName[ClassCounter]&={Name}
         InstanceMethodSource[ClassCounter]&=ClassCounter
         InstanceMethodID[ClassCounter]&=ID
      else
         ClassMethodName[ClassCounter]&={Name}
         ClassMethodSource[ClassCounter]&=ClassCounter
         ClassMethodID[ClassCounter]&=ID
      end if
   else
      if entityType=INSTANCE then
         methSource=InstanceMethodSource[ClassCounter]
      else
         methSource=ClassMethodSource[ClassCounter]
      end if
      if methSource[index]=ClassCounter then
         fatal_error("In method().  "&CONSTANT_CAP_NAME[entityType]
         &" method "&Name&" has already been defined.")
      end if
      if entityType=INSTANCE then
         InstanceMethodSource[ClassCounter][index]=ClassCounter
         InstanceMethodID[ClassCounter][index]=ID
      else
         ClassMethodSource[ClassCounter][index]=ClassCounter
         ClassMethodID[ClassCounter][index]=ID
      end if
   end if
end procedure

-- Exception handling:

global procedure throw(object theException)
   integer thisClass
   if Context[ProgramCounter]=CLASS_DEFINITION then
      fatal_error("In throw().  Not allowed.")
   end if
   if not class_entity(theException) then
      fatal_error("In throw().  Invalid exception.")
   end if
   thisClass=theException[1]
   if ClassType[thisClass]!=EXCEPTION then
      fatal_error("In throw().  Class "&get_class_name(thisClass)
      &" is not an exception.")
   end if
   if PendingException!=NIL then
      fatal_error("In throw().  Exception "&get_class_name(thisClass)
      &" thrown.")
   end if
   PendingException=thisClass
   CaughtException=NIL
end procedure

global function catch(object theException)
   integer thisClass
   if Context[ProgramCounter]=CLASS_DEFINITION then
      fatal_error("In catch().  Not allowed.")
   end if
   if not class_entity(theException) then
      fatal_error("In catch().  Invalid exception.")
   end if
   thisClass=theException[1]
   if ClassType[thisClass]!=EXCEPTION then
      fatal_error("In catch().  Class "&get_class_name(thisClass)
      &" is not an exception.")
   end if
   if PendingException=NIL then return FALSE end if
   if not find(thisClass,SuperclassList[PendingException]) then
      return FALSE
   end if
   CaughtException=PendingException
   PendingException=NIL
   return TRUE
end function

global function caught()
   if Context[ProgramCounter]=CLASS_DEFINITION then
      fatal_error("In caught().  Not allowed.")
   end if
   if CaughtException=NIL then
      return Null_Instance
   end if
   return {CaughtException,0,MARKER}
end function

global function success()
   if Context[ProgramCounter]=CLASS_DEFINITION then
      fatal_error("In success().  Not allowed.")
   end if
   return PendingException=NIL
end function 


                                                             
-- Predefined class methods: 


ClassCounter=1
ProgramCounter=2
Context[2]=CLASS_DEFINITION
Caller[2]=1
This[2]=Entity
ThisMethod[2]=NIL   -- class()

function Entity_new()
   integer thisClass
   sequence thisEntity
   thisEntity=This[ProgramCounter]
   thisClass=thisEntity[1]
   InstanceCounter+=1
   if InstanceCounter>length(InstanceEntity) then
      InstanceEntity&=repeat(NIL,length(InstanceEntity))
   end if
   InstanceEntity[InstanceCounter]=InstancePropertyDefault[thisClass]
   thisEntity[2]=InstanceCounter
   return thisEntity
end function

method("new",0,CLASS,routine_id("Entity_new")) 


function Entity_clone() 
   sequence thisEntity
   thisEntity=This[ProgramCounter]
   InstanceCounter+=1
   if InstanceCounter>length(InstanceEntity) then
      InstanceEntity&=repeat(NIL,length(InstanceEntity))
   end if
   InstanceEntity[InstanceCounter]=InstanceEntity[thisEntity[2]]
   thisEntity[2]=InstanceCounter
   return thisEntity
end function

method("clone",0,INSTANCE,routine_id("Entity_clone"))


function Entity_delete()
   integer thisInstance
   sequence thisEntity
   thisEntity=This[ProgramCounter]
   thisInstance=thisEntity[2]
   InstanceEntity[thisInstance]=NIL
   return Null_Instance
end function

method("delete",0,INSTANCE,routine_id("Entity_delete"))


ProgramCounter=1  -- end_class()
ClassCounter=3



